############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import urlparse, time, types
from xml.dom.ext.reader import PyExpat
from xml.ns import WSDL
from Ft.Xml import XPath

from pyGridWare.ogsi.services.OGSI_services_types import ns1 as ogsi
from pyGridWare.ogsi.services import ServiceException

from pyGridWare import getLogger
logger = getLogger(__name__)


""" GridService

    This module defines some Base classes for services.

    classes:
        GridServiceException
          -- exception
        Reference
          -- contains a singleton class represents the service
             instance node tree.
        GridServicePort
          -- cans basic gridService port methods.
        PortType
          -- GWSDL PortType class, defines ServiceData class, etcetera.
        GridService_portType
          -- portType interface, non-implemented callbacks.
        GridService_gPortType
          -- gwsdl portType instance, contains definitions for serviceData.
"""

class GridServiceException(ServiceException): pass

class Reference:
    class NodeTree:
        """Simple dictionary implementation of a node tree
        """
        def __init__(self):
            self.__dict = {}
            self.post_dict = {}

        def __str__(self):
            return str(self.__dict)

        def getNode(self, path):
            node = self.__dict
            for i in path.split('/'):
                if not i: continue
                node = node[i]
            return node

        def setNode(self, pyobj, path):
            self.post_dict[path] = pyobj
            node = self.__dict
            dir = path.split('/')
            for i in dir[:-1]:
                if not i: continue
                if not node.has_key(i): node[i] = {}
                node = node[i]
            node[dir[-1]] = pyobj
    node_tree = NodeTree()

    def getDict(self):
        return Reference.node_tree.post_dict

    def getNode(self, path):
        return Reference.node_tree.getNode(path)

    def setNode(self, pyobj, path):
        Reference.node_tree.setNode(pyobj, path)


class Binding:
    soapAction = []

    def __init__(self):
        return

    def __call___(self, bind, msg):
        op = self.getOperation(soapAction=bind)
        method = getattr(self, op)
        return method(port=self.__port, bind=self.__bind, msg=msg)

    def getOperation(self, soapAction):
        """soapAction --
           return -- callback method name.
        """
        return self.soapAction.get(soapAction)


class Port:
    def __init__(self):
        return


class Service(Reference):
    """Service must define callback methods that delegate to a port.
       The callback methods in services can delegate to multiple ports,
       or alternative ports as long as they have different bindings.

       All Service callback methods must be defined like:
           (port, bind, *args)

       Callbacks must be retrieved with a binding value (soapAction)
       via the "getCallBack" or "call" methods.
    """
    class __Method:
        """Binds the port to the method call by adding a few parameters
           to the callback function.
        """
        def __init__(self, call, port, bind):
            """call -- bound service callback method.
               port -- port to delegate to.
               bind -- soapAction value.
            """
            self.__call = call
            self.__port = port
            self.__bind = bind

        def __call__(self, msg):
            return self.__call(port=self.__port, bind=self.__bind, msg=msg)

    def __init__(self):
        self.ports = []

    def __call__(self, bind, msg):
        """Delegation, port and soapAction set in __Method instance.
           Allows multiple ports to declare the same callback method, but 
           depends on unique binding values.

           bind -- soapAction value
           msg -- Input Message to the callback.
           return -- from the service method, pyobj representing a 
               messsage/typecode.
        """
        method = self.getCallBack(bind)
        return method(self, msg)

    def getCallBack(self, bind):
        """Returns port and service callback method.
           bind -- soapAction value
        """
        for p in self.ports:
            op = p.getOperation(soapAction=bind)
            if op: break
        else:
            raise GridServiceException, 'Service Missing method %s' %bind
        return self.__Method(call=getattr(self,op),port=p,bind=bind)

    def getWSDLReference(self):
        logger.info('GridServiceInterface.getWSDLReference')
        return self.getWSDLReferenceDOM(self.reference)

    def getWSDLReferenceDOM(self, wsdlPath):
        """Returns a DOM from a file path to a WSDL service description.
           <?xml version="1.0" encoding="UTF-8"?>
        """
        st = open(wsdlPath)
        reader = PyExpat.Reader()
        document = reader.fromStream(st)
        processorNss = {'wsdl':WSDL.BASE, 'soap':WSDL.BIND_SOAP}
        context = XPath.Context.Context(document, processorNss=processorNss)
        location = 'location'
        expr = 'wsdl:definitions//soap:*[@%s]' %location
        compExpr = XPath.Compile(expr)
        for e in compExpr.evaluate(context):
            scheme,netloc,path,params,query,fragment = \
                urlparse.urlparse(e.getAttribute(location))
            url_tuple = (scheme,
                         '%s:%d' %(self.addr,self.port),
                         self.post,
                         params,
                         query,
                         fragment)
            url = urlparse.urlunparse(url_tuple)
            e.setAttribute(location, url)
        expr = 'wsdl:definitions/wsdl:import[@%s]' %location
        compExpr = XPath.Compile(expr)
        for e in compExpr.evaluate(context):
            scheme,netloc,path,params,query,fragment = \
                urlparse.urlparse(e.getAttribute(location))
            if not scheme:
                #support relative paths...
                scheme = 'http'
                endIdx = wsdlPath.rfind('/')
                if endIdx > 0:
                    abs_path = '%s/%s' %(wsdlPath[:endIdx],path)
                else:
                    abs_path = path
            url_tuple = (scheme or 'http',
                         '%s:%d' %(self.addr,self.port),
                         abs_path,
                         params,
                         query,
                         fragment)
            url = urlparse.urlunparse(url_tuple)
            e.setAttribute(location, url)
        return document


class PortType:
    """
  <complexType name="PortType">
    <sequence>
      <element name="operation" minOccurs="0" maxOccurs="unbounded">
        <complexType>
          <sequence>
            <any namespace="##any"/>
          </sequence>
        </complexType>
      </element>
      <element name="serviceData" minOccurs="0" maxOccurs="unbounded" type="sd:ServiceDataType"/>
      <element name="staticServiceDataValues" 
               type="tns:StaticServiceDataValues"
               minOccurs="0" 
               maxOccurs="1"/>
    </sequence>
  </complexType>

  <complexType name="StaticServiceDataValues">
    <sequence>
      <any namespace="##any" minOccurs="0" maxOccurs="unbounded"/>
    </sequence>
  </complexType>
    """
    class ServiceDataType:
        """
        <complexType name="ServiceDataType">
          <sequence>
            <any namespace="##any" minOccurs="0" maxOccurs="unbounded"/>
          </sequence>
          <attribute name="name" type="NMTOKEN"/>
          <attribute name="type" type="QName"/>
          <attribute name="minOccurs" type="nonNegativeInteger"/>
          <attribute name="maxOccurs" type="string"/>
          <attribute name="mutability" type="tns:MutabilityType"/>
          <attribute name="modifiable" type="boolean" default="false"/>
          <attribute name="nillable" type="boolean" default="false"/>
        </complexType>
        """
        pass

    class ServiceData(ServiceDataType):
        name=None
        type=None
        minOccurs=1
        maxOccurs="unbounded"
        mutability=None
        modifiable=False
        nillable=False

        def __init__(self):
            self.__modificationTime = time.gmtime(time.time())
            PortType.ServiceDataType.__init__(self)
            self.__instances = []

        def addInstance(self, pyclass=None, name=None):
            """append a service data instance, unless only one 
               instance is allowed.

               pyclass - allow derived type to be specified.
            """
            if pyclass and not issubclass(pyclass, self.type):
                raise TypeError, 'bad type %s, must subclass %s' \
                    %(str(pyclass), str(self.type))
            elif not pyclass:
                pyclass = self.type
            if not name:
                name = self.name

            if (not isinstance(self.maxOccurs, types.StringType)) and\
                self.maxOccurs >= len(self.__instances):
                 raise GridServiceException, 'SDE %s filled' %name
            if self.minOccurs == 1 and\
                (not isinstance(self.maxOccurs, types.StringType)) and\
                self.maxOccurs == 1:
                self.__instances = (pyclass(pname=name),)
            else:
                self.__instances.append(pyclass(pname=name))

        def getInstances(self):
            """returns instances of service data, return a list 
               unless only one instance is allowed.
            """
            if self.minOccurs == 1 and\
                (not isinstance(self.maxOccurs, types.StringType)) and\
                self.maxOccurs == 1:
                return self.__instances[0]
            return self.__instances

    def __init__(self):
        self.serviceDataValues = {}
        for name,sd in self.serviceDataTypes.items():
            self.serviceDataValues[name] = sd()


class GridService_portType:
    def __init__(self, addr, port, post, reference):
        self.addr = addr
        self.port = port
        self.post = post
        self.reference = reference

    def setServiceData(self, msg):
        raise NotImplementedError, 'portType is an interface'

    def destroy(self, msg):
        raise NotImplementedError, 'portType is an interface'

    def findServiceData(self, msg):
        raise NotImplementedError, 'portType is an interface'

    def requestTerminationBefore(self, msg):
        raise NotImplementedError, 'portType is an interface'

    def requestTerminationAfter(self, msg):
        raise NotImplementedError, 'portType is an interface'


class GridService_gPortType(PortType, GridService_portType):
    """
    <gwsdl:portType name="GridService">
        <operation name="findServiceData">
        <operation name="setServiceData">
        <operation name="requestTerminationAfter">
        <operation name="requestTerminationBefore">
        <operation name="destroy">

        <sd:staticServiceDataValues>
               <ogsi:findServiceDataExtensibility inputElement="ogsi:queryByServiceDataNames"/>
               <ogsi:setServiceDataExtensibility inputElement="ogsi:setByServiceDataNames"/>
               <ogsi:setServiceDataExtensibility inputElement="ogsi:deleteByServiceDataNames"/>
        </sd:staticServiceDataValues>
    """
    serviceDataTypes = {}
    staticServiceData = {}
    class interface(PortType.ServiceData):
        """The QNames of all the portType(s) within the service instance's complete interface 
           definition. This set MUST contain the transitive closure of the portType QNames 
           implemented by the service. 
        """
        name="interface"
        type="xsd:QName"
        minOccurs=1
        maxOccurs="unbounded"
        mutability="constant"
        modifiable=False
        nillable=False
    serviceDataTypes[interface.name] = interface

    class serviceDataName(PortType.ServiceData):
        """The QNames for each service data element supported by this service instance. This set 
           MUST contain the QNames of all service data elements declared in the WSDL definition 
           or the portType. This set MAY also contain the QNames of service data elements added 
           dynamically by the instance. 
        """
        name="serviceDataName"
        type="xsd:QName"
        minOccurs=0
        maxOccurs="unbounded"
        mutability="mutable"
        modifiable=False
        nillable=False
    serviceDataTypes[serviceDataName.name] = serviceDataName

    class factoryLocator(PortType.ServiceData):
        """A service locator to the factory that created this Grid service instance. If the instance was 
           not created by a factory, this value MUST be xsi:nil. This locator MUST refer to the 
           same factory service throughout the lifetime of this Grid service instance, although the 
           handles and references contained in this locator MAY change during the service 
           instance's lifetime.  
        """
        name="factoryLocator"
        type='ogsi.LocatorType'
        minOccurs=1
        maxOccurs=1
        mutability="mutable"
        modifiable=False
        nillable=True
    serviceDataTypes[factoryLocator.name] = factoryLocator

    class gridServiceHandle(PortType.ServiceData):
        """Zero or more Grid Service Handles of this Grid service instance. There MAY be handles 
           to this service instance that are not included in this set. 
        """
        name="gridServiceHandle"
        type='ogsi.HandleType'
        minOccurs=0
        maxOccurs="unbounded"
        mutability="extendable"
        modifiable=False
        nillable=False
    serviceDataTypes[gridServiceHandle.name] = gridServiceHandle

    class gridServiceReference(PortType.ServiceData):
        """One or more Grid Service References to this Grid service instance. One service data 
           value element MUST be the WSDL representation of the GSR. Other service data value 
           elements may represent other forms of the GSR. There MAY be references to this service 
           instance that are not included in this set. 
        """
        name="gridServiceReference"
        type='ogsi.ReferenceType_Def'
        minOccurs=1
        maxOccurs="unbounded"
        mutability="mutable"
        modifiable=False
        nillable=False
    serviceDataTypes[gridServiceReference.name] = gridServiceReference

    class findServiceDataExtensibility(PortType.ServiceData):
        """A set of operation extensibility declarations (§7.8) for the findServiceData operation. 
           Any conforming inputElement declared by values of this SDE MAY be used by the client 
           as a QueryExpression parameter to the instance's findServiceData operation, and implies 
           the query semantics and return values that may result from the query. 
        """
        name="findServiceDataExtensibility"
        type='ogsi.OperationExtensibilityType_Def'
        minOccurs=1
        maxOccurs="unbounded"
        mutability="static"
        modifiable=False
        nillable=False
        def __init__(self):
            sd = self.addInstance()
            sd.setAttribute_inputElement('queryByServiceDataNames')
    serviceDataTypes[findServiceDataExtensibility.name] = findServiceDataExtensibility

    class setServiceDataExtensibility(PortType.ServiceData):
        """A set of operation extensibility declarations (7.8) for the setServiceData operation. Any 
           conforming inputElement declared by values of this SDE MAY be used by the client as 
           the UpdateExpression parameter to the instance's setServiceData operation, and implies 
           the update semantics and return values that may result from the update. 
        """
        name="setServiceDataExtensibility"
        type='ogsi.OperationExtensibilityType_Def'
        minOccurs=1
        maxOccurs="unbounded"
        mutability="static"
        modifiable=False
        nillable=False
        def __init__(self):
            sd = self.addInstance()
            sd.setAttribute_inputElement('setByServiceDataNames')
            sd = self.addInstance()
            sd.setAttribute_inputElement('deleteByServiceDataNames')

        def evaluate(self, updateExpression, serviceData):
            for sd in self.getInstances():
                inputElement = sd.attributes.get('inputElement')
                elementName = '_%s' %inputElement
                if hasattr(updateExpression, elementName):
                    method = getattr(self, inputElement)
                    pyobj = getattr(updateExpression, elementName)
                    break
            else:
                #ExtensibilityNotSupportedFault
                raise GridServiceException, 'bad inputElement declaration %s in updateElement'\
                    %str(updateElement)
            return method(pyobj=pyobj, serviceData=serviceData)

        def setByServiceDataNames(self, pyobj, serviceData): 
            """A setByServiceDataNames results in the update of the SDE values in the serviceDataValues 
               containing the service data elements listed in the setByServiceDataNames expression. The names 
               listed in the expression MUST be among the serviceDataNames (see §9.1) contained in this 
               service instance with an attribute of modifiable="true." There is no guarantee that the update is 
               executed in any particular sequence. It is the service's responsibility to ensure that the SDE 
               values are successfully updated in the underlying service's state. The service SHOULD update as 
               many SDEs successfully as it can and return those that failed. The service MAY return after the 
               first failure and not continue.   

               The setServiceData operation's Result output parameter for a setByServiceDataNames invocation 
               MUST be a serviceDataValues element containing the service data elements listed in the 
               setByServiceDataNames expression that failed to be replaced. An empty value set indicates 
               success. 
            """
            if not isinstance(pyobj, ogsi.setByServiceDataNames_Dec):
                #ExtensibilityNotSupportedFault: Indicates that the service cannot evaluate the 
                #      UpdateExpression because its type is not supported by this service. 
                raise TypeError, 'service cant evaluate UpdateExpression %s, %s'\
                    %(str(pyobj.__class__),str(ogsi.setByServiceDataNames_Dec))

        def deleteByServiceDataNames(self, pyobj, serviceData):
            """A deleteByServiceDataNames results in the deletion of all SDE values for the service data 
               elements listed in the deleteByServiceDataNames expression. The names listed in the expression 
               MUST be among the serviceDataNames (see §9.1) contained in this service instance with an 
               attribute of modifiable="true". There is no guarantee that the update is executed in any particular 
               sequence. It is the service's responsibility to ensure that the SDEs are successfully deleted in the 
               underlying service's state. The service SHOULD delete as many SDEs successfully as it can and 
               return those that failed. The service MAY return after the first failure and not continue. 

               The setServiceData operation's Result output parameter for a deleteByServiceDataNames 
               invocation MUST be a serviceDataValues element containing the service data elements listed in 
               the deleteByServiceDataNames expression that failed to be deleted. An empty value set indicates 
               success. 
            """
            if not isinstance(pyobj, ogsi.deleteByServiceDataNames_Dec):
                raise TypeError, 'bad type %s, expecting %s'\
                    %(str(pyobj.__class__),str(ogsi.deleteByServiceDataNames_Dec))


    serviceDataTypes[setServiceDataExtensibility.name] = setServiceDataExtensibility

    class terminationTime(PortType.ServiceData):
        """The termination time for this service (see §7.6).
        """
        name="terminationTime"
        type= ogsi.TerminationTimeType_Def
        minOccurs=1
        maxOccurs=1
        mutability="mutable"
        modifiable=False
        nillable=False
    serviceDataTypes[terminationTime.name] = terminationTime


    def __init__(self, addr, port, post, reference):
        PortType.__init__(self)
        GridService_portType.__init__(self, addr, port, post, reference)

    #operation
    def setServiceData_SDE(self, updateExpression):
        """This operation allows for the modification of a service data element's values, if its service data 
           declaration specifies modifiable="true". Changing a modifiable service data element implies 
           changing the corresponding state in the underlying service instance. If no service data elements 
           have a modifiable="true" attribute then setServiceData is essentially disabled. 

           Input 
              -UpdateExpression: The update to be performed. This extensible parameter MUST 
                  conform to an inputElement declaration denoted by one of the 
                  setServiceDataExtensibility SDE values. Note: The service infers what to do based on the 
                  tag of the root element of this argument 

           Output 
              -Result: The result of the update. The format of this result is dependent upon the 
                  UpdateExpression. 

           Faults 
              -ExtensibilityNotSupportedFault: Indicates that the service cannot evaluate the 
                  UpdateExpression because its type is not supported by this service. 
              -ExtensibilityTypeFault: Indicates that the value passed as the UpdateExpression violates 
                  that value's type. 
              -CardinalityViolationFault: Indicates that operation requested would violate the 
                  "minOccurs" and/or "maxOccurs" attributes of the service's SDE(s). 
              -MutabilityViolationFault: Indicates that the UpdateExpression was not consistent with 
                  the "mutability" attribute of the service's SDE(s). 
              -ModifiabilityViolationFault: Indicates that the UpdateExpression was not consistent with 
                  the "modifiable" attribute of the services's SDE(s).  
              -TypeViolationFault: Indicates that the UpdateExpression contains values that do not 
                  conform to the XSD type of the service's SDE(s). 
              -IncorrectValueFault: Indicates that the UpdateExpression contains values that are XSD 
                  type conformant, but are not acceptable to the service for other reasons. 
              -PartialFailureFault: Indicates that service was unable to satisfy all portions of the 
                  UpdateExpression.  This fault extends FaultType with an element that contains a list of 
                  qnames of SDEs from the UpdateExpression that could not be updated.  This fault MAY 
                  have 1 or more "faultcause" elements that describe in more detail the portions that failed, 
                  using any of the fault types allowed by setServiceData.  
              -Fault: Any other fault. 
        """
        sdName = self.setServiceDataExtensibility.name
        sd = self.serviceDataValues[sdName]
        return sd.evaluate(updateExpression=updateExpression, serviceData=self.serviceDataValues)

    #setServiceData --> setByServiceDataNames
    def setByServiceDataNames_SDE(self, pyobj): 
        """A setByServiceDataNames results in the update of the SDE values in the serviceDataValues 
           containing the service data elements listed in the setByServiceDataNames expression. The names 
           listed in the expression MUST be among the serviceDataNames (see §9.1) contained in this 
           service instance with an attribute of modifiable="true." There is no guarantee that the update is 
           executed in any particular sequence. It is the service's responsibility to ensure that the SDE 
           values are successfully updated in the underlying service's state. The service SHOULD update as 
           many SDEs successfully as it can and return those that failed. The service MAY return after the 
           first failure and not continue.   

           Partial failures (i.e. some SDEs are updated but some are not) MUST be indicated by returning a 
           PartialFailureFault, with a list of the QNames of the SDEs that could not be set.  This fault MAY 
           also have one "faultcause" element for each SDE in the setByServiceDataNames expression that 
           could not be set, where the type of the "faultcause" is any of the valid setServiceData faults, and 
           the extensibility element in the "faultcause" is the SDE value from the setByServiceDataNames 
           expression that could not be set. 

           A setByServiceDataNames MUST adhere to the mutability attribute of the SDEs as specified in 
           §6.2.3. If the mutability value is "static" or "constant," then setByServiceDataNames is not 
           allowed. If the mutability value is "extendable," then setByServiceDataNames MUST append to 
           the new elements to the SDE's existing values. If the mutability value is "mutable,"then 
           setByServiceDataNames MUST replace the existing SDE values with the ones that are passed in. 
           The setByServiceDataNames MUST, for each SDE named in the expression, result in the 
           serviceData element values adhering to the minOccurs/maxOccurs rules defined in the 
           serviceData declaration. 

           The setServiceData operation's Result output parameter for a setByServiceDataNames invocation 
           MUST be a serviceDataValues element containing the service data elements listed in the 
           setByServiceDataNames expression that failed to be replaced. An empty value set indicates 
           success. 

           Note that the client SHOULD NOT assume that the values in the service's serviceDataValues are 
           in any way coherent with each other. Clients also SHOULD NOT assume that the replacing of 
           any service data elements is carried out with any coherency guarantee. 
        """
        return

    #setServiceData -> deleteByServiceDataNames
    def deleteByServiceDataNames_SDE(self, pyobj):
        """A deleteByServiceDataNames results in the deletion of all SDE values for the service data 
           elements listed in the deleteByServiceDataNames expression. The names listed in the expression 
           MUST be among the serviceDataNames (see §9.1) contained in this service instance with an 
           attribute of modifiable="true". There is no guarantee that the update is executed in any particular 
           sequence. It is the service's responsibility to ensure that the SDEs are successfully deleted in the 
           underlying service's state. The service SHOULD delete as many SDEs successfully as it can and 
           return those that failed. The service MAY return after the first failure and not continue. 

           Partial failures (i.e. some SDEs are deleted but some are not) MUST be indicated by returning a 
           PartialFailureFault, with a list of the QNames of the SDEs that could not be deleted.  This fault 
           MAY also have one "faultcause" element for each SDE in the deleteByServiceDataNames 
           expression that could not be set, where the type of the "faultcause" is any of the valid 
           deleteServiceData faults, and the extensibility element in the "faultcause" contains the QName 
           that could not be deleted. 

           A deleteByServiceDataNames MUST adhere to the mutability attribute of the SDE as specifie d in 
           §6.2.3. If the mutability value is "static," "constant," or "extendable," then 
           deleteByServiceDataNames is not allowed. If the mutability value is "mutable," then 
           deleteByServiceDataNames will delete all elements with the SDE names. The 
           deleteByServiceDataNames MUST, for each SDE named in the expression, result in the 
           serviceData element values adhering to the minOccurs/maxOccurs rules defined in the 
           serviceData declaration. 

           The setServiceData operation's Result output parameter for a deleteByServiceDataNames 
           invocation MUST be a serviceDataValues element containing the service data elements listed in 
           the deleteByServiceDataNames expression that failed to be deleted. An empty value set indicates 
           success. 

           Note that the client SHOULD NOT assume that the values in the service's serviceDataValues are 
           in any way coherent with each other before they are updated. Clients also SHOULD NOT assume 
           that the deletions of any service data elements are carried out with any coherency guarantee. 
        """
        if not isinstance(pyobj, ogsi.deleteByServiceDataNames_Dec):
            raise TypeError, 'bad type %s, expecting %s'\
                %(str(pyobj.__class__),str(ogsi.deleteByServiceDataNames_Dec))

    #operation
    def destroy_SDE(self, pyobj):
        """Explicitly request destruction of this service. Upon receipt of an explicit destruction request, a 
           Grid service MUST either initiate its own destruction and return a response acknowledging the 
           receipt of the destroy message; or ignore the request and return a fault message indicating failure. 

           Once destruction of the Grid service is initiated, the service SHOULD NOT respond successfully 
           to further requests. Following a successful destroy operation, the client SHOULD NOT rely on 
           the existence of the service. 

           Input: 
               -None (An empty input message.) 
           Output: 
               -None (An empty output message, acknowledging that the destroy has been initiated.) 
           Fault(s): 
               -ServiceNotDestroyedFault: Indicates that the service chose not to initiate self-destruction. 
               -Fault: Any other fault.  
        """
        return

    #operation
    def findServiceData_SDE(self, queryExpression):
        """Query the service data. 
           Input 
              -QueryExpression: The query to be performed. This extensible parameter MUST conform 
                  to an inputElement declaration denoted by one of the findServiceDataExtensibility SDE 
                  values. Note: The service infers what to do based on the tag of the root element of this 
                  argument. 
           Output 
              -Result: The result of the query. The format of this result is dependent upon the 
                  QueryExpression. 

           Faults
              -ExtensibilityNotSupportedFault: Indicates that the service cannot evaluate the 
                  QueryExpression because its type is not supported by this service. 
              -ExtensibilityTypeFault: Indicates that the value passed as the QueryExpression violates 
                  that value's type. 
              -TargetInvalidFault: Indicates that one or more of the SDEs that the QueryExpression 
                  requires do not exist in this service. 
              -Fault: Any other fault. 

        """
        logger.info('GridService_gPortType.findServiceData')
        sdName = None

        if hasattr(queryExpression, '_queryByServiceDataNames'):
            sdName = queryExpression._queryByServiceDataNames._name
        else:
            logger.error('Unsupported queryExpression %s', str(queryExpression))
            raise GridServiceException, 'Unsupported queryExpression'

        pyobj = None
        for func in self.serviceDataValues.keys():
            var = '_%s' %func
            if hasattr(queryExpression, var):
                pyobj = getattr(self, func)(pyobj=getattr(queryExpression,var))
                break
        else:
            logger.info('check Dynamic serviceData, Not implemented')
        return pyobj

    #findServiceData --> queryByServiceDataNames
    def queryByServiceDataNames_SDE(self, pyobj):
        """A queryByServiceDataNames results in a serviceDataValues element containing the service data 
           elements named in the queryByServiceDataNames parameter. The names listed in the 
           queryByServiceDataNames MUST be contained in the serviceDataName service data element 
           (see §9.1) contained in this service instance. 

           The findServiceData operation's Result output parameter for a queryByServiceDataNames query 
           MUST be a sd:serviceDataValues element containing the service data elements listed in the 
           queryByServiceDataNames.  
        """
        logger.info('GridService_gPortType.queryByServiceDataNames')
        if not isinstance(pyobj, ogsi.queryByServiceDataNames_Dec):
            raise TypeError, 'bad type %s, expecting %s'\
                %(str(pyobj.__class__),str(ogsi.queryByServiceDataNames_Dec))
        sd = []
        for name in query._queryByServiceDataNames._name:
            if self.serviceDataValues.has_key(name):
                sd.append(self.serviceDataValues[name].getInstances())
            else:
                logger.error('queryByServiceDataNames missing name(%s)', name)
        return sd

    #operation 
    def requestTerminationBefore_SDE(self, terminationTime):
        """Request that the termination time of this service be changed. The request specifies the latest 
           desired termination time. Upon receipt of the request, MAY adjust its termination time, if 
           necessary, based on its own polices and the requested time. Upon receipt of the response, the 
           client SHOULD discard any responses that have arrived out of order, based on the timestamp in 
           the response. 

           Input: 
               -TerminationTime: The latest termination time of the Grid service that is acceptable to the 
                   client. Its type is an ExtendedDateTimeType as defined in §7.2. A time in the past 
                   indicates a desire that the service terminate as soon as possible. The special value 
                   "infinity" indicates that the clie nt no longer cares about a maximum termination time. 

           Output: 
               -CurrentTerminationTime: An element of type TerminationTimeType (as defined in §7.6) 
                   that gives the service's currently planned termination time.  The timestamp attribute of 
                   this CurrentTerminationTime MUST be the time at which the service handled this 
                   request. 
           Fault(s): 
               -TerminationTimeUnchangedFault: Indicates that the service chose to ignore the 
                   requested termination time. 
               -Fault: Any other fault. 

        """
        pyobj = self.serviceDataValues[terminationTime.name].getInstances()[0]
        pyobj.setAttribute_before(terminationTime)
        
    #operation
    def requestTerminationAfter_SDE(self, terminationTime):
        """Request that the termination time of this service be changed. The request specifies the earliest 
           desired termination time. Upon receipt of the request, the service MAY adjust its termination 
           time, if necessary, based on its own polices and the requested time. Upon receipt of the response, 
           the client SHOULD discard any responses that have arrived out of order, based on the 
           CurrentTimestamp in the response. 

           Input: 
               -TerminationTime: The earliest termination time of the Grid service that is acceptable to 
                   the client. Its type is an ExtendedDateTimeType as defined in §7.2. A value in the past 
                   indicates that the client no longer cares about the earliest termination time. The special 
                   value "infinity" means that the client requests that the service continue to exist 
                   indefinitely. 
           Output: 
              -CurrentTerminationTime: An element of type TerminationTimeType (as defined in §7.6) 
                   that gives the service's currently planned termination time.  The timestamp attribute of 
                   this CurrentTerminationTime MUST be the time at which the service handled this 
                   request. 
           Fault(s): 
              -TerminationTimeUnchangedFault: Indicates that the service chose to ignore the 
                   requested termination time. 
              -Fault: Any other fault. 
        """
        pyobj = self.serviceDataValues[terminationTime.name].getInstances()[0]
        pyobj.setAttribute_after(terminationTime)



'''
class GridServiceInteface:
    """Example implementation for a GridService port.
       Define set/get operations for all serviceData defined by
       GridService_portType.  These are methods that a service can 
       use to managage its ports.  

       serviceData:
           gridServiceReference
           findServiceDataExtensibility
           setServiceDataExtensibility
           terminationTime
    """
    def __init__(self):
        return

    def setTerminationTime(self, after=None, before=None):
        logger.info('GridServiceInterface.setTerminationTime')
        self.serviceData[GridService_portType.terminationTime.name].addInstance()
        pyobj = self.serviceData[GridService_portType.terminationTime.name].getInstances()
        if not after and not before:
            pyobj.setAttribute_before(time.gmtime(time.time()))
        if after: pyobj.setAttribute_after(time.struct_time(after))
        if before: pyobj.setAttribute_before(time.struct_time(before))

    def getTerminationTime(self):
        logger.info('GridServiceInterface.getTerminationTime')
        return self.serviceData[GridService_portType.terminationTime.name].getInstances()

    def setWSDLReference(self):
        logger.info('GridServiceInterface.setWSDLReference')
        self.serviceData[GridService_portType.gridServiceReference.name].addInstance(\
            pyclass=ogsi.WSDLReferenceType_Def, name='reference')
        pyobj = self.serviceData[GridService_portType.gridServiceReference.name].getInstances()[-1]
        pyobj._dom = self.getWSDLReferenceDOM(wsdlPath=self.reference)

'''
